#include "stdafx.h"
#include "rootfilter.h"
#include <cassert>

using namespace std;

RootFilter rootFilter;

ItemInfo RootFilter::parDescr[] = 
{
    {
        TYPE_CHAR,                                      // array of chars
        MPX_MAX_PATH,                                   // max. size of array
        ITEMINFO_CANCHANGE,                             // flag that means user can change this value
        "FilePath",                                     // name of parameter
        "File path name (e.g. C:\\data\\test.root)",    // description of parameter
        0                                               // data (we do not need it here)
    },
};

int RootFilter::filterFunc(INTPTR userData, byte *in, byte *out)
{
    // extract frame ID and instance ID from *in pointer
    return rootFilter.filter(*(FRAMEID *) in, *(ITEMID *) (in + sizeof(FRAMEID)));
}

int RootFilter::createInstFunc(INTPTR userData, ITEMID *instID)
{
    return rootFilter.createInst(instID);
}

int RootFilter::deleteInstFunc(INTPTR userData, ITEMID instID)
{
    return rootFilter.deleteInst(instID);
}

int RootFilter::setInstParFunc(INTPTR userData, ITEMID instID, int idx, byte *data, int dataSize)
{
    return rootFilter.setInstPar(instID, idx, data, dataSize);
}

int RootFilter::getInstParFunc(INTPTR userData, ITEMID instID, int idx, ItemInfo *info, int *dataSize)
{
    return rootFilter.getInstPar(instID, idx, info, dataSize);
}

RootFilter::RootFilter()
{
    InitializeCriticalSection(&critSection);
}

RootFilter::~RootFilter()
{
    for (unsigned i = 0; i < instCfgs.size(); i++)
        delete instCfgs[i];
    instCfgs.clear();
    DeleteCriticalSection(&critSection);
}

// initialization of filter (registration of filter function)
void RootFilter::init()
{
    // fill structure containing interface to adjustable filter
    AdjFilterPars adjFiltPars = 
    {
        0, 0, 0, 0, 0,              // we do not use User Data here
        createInstFunc,             // create instance function
        deleteInstFunc,             // delete instance function
        setInstParFunc,             // set instance par.
        getInstParFunc,             // get instance par.
    };

    // fill structure describing function
    ExtFunctionInfo efi = 
    {
        filterFunc,                                 // pointer to function performing the action
        2,                                          // number of input parameters
        0,                                          // number of output parameters
        {
            {TYPE_U32, 1, "ID of frame"},           // description of first parameter
            {TYPE_U32, 1, "ID of filter instance"}, // description of second parameter
        },
        {                                           // empty array of output parameter
        },
        PLUGIN_NAME,                                // name of plugin
        "RootFilter",                               // name of function
        "Filter which saves frame in root format",  // descr. of function
        (INTPTR) &adjFiltPars,                      // user data has to contain pointer to AdjFilterPars structure
        FUNCFLAG_FILTER | FUNCFLAG_ADJFILTER        // flags which tells manager that this function is filter
    };
    mgr->addFuncItem(&efi);
}

int RootFilter::createInst(ITEMID *instID)
{
    InstCfg *cfg = new InstCfg;
    LockCritSect lock(&critSection);
    *instID = instCfgs.size();
    instCfgs.push_back(cfg);
    return 0;
}

int RootFilter::deleteInst(ITEMID instID)
{
    LockCritSect lock(&critSection);
    assert(instID < (int) instCfgs.size());
    delete instCfgs[instID];
    instCfgs[instID] = 0;
    return 0;
}

// macro which checks inst. ID and gets inst. cfg from ID
#define GETINST(id) \
    LockCritSect lock(&critSection);\
    if (instID >= (int) instCfgs.size())\
        return -1;\
    InstCfg *cfg = instCfgs[instID];\
    if (!cfg)\
        return -1;

int RootFilter::setInstPar(ITEMID instID, int idx, byte *data, int dataSize)
{
    GETINST(instID);
    if ((unsigned) idx >= parCount)
        return -1;
    assert(data);
    // we have only one parameter, so set it
    cfg->filePathName = string((const char *)data, dataSize);       // we do not know if data contains zero terminated string so we use dataSize
    return 0;
}

int RootFilter::getInstPar(ITEMID instID, int idx, ItemInfo *info, int *dataSize)
{
    // if idx is negative, write number of adjustable parameters to dataSize
    if (idx < 0)
    {
        *dataSize = parCount;
        return 0;
    }
    GETINST(instID);
    if (idx >= parCount)
        return -1;

    // fill ItemInfo structure
    assert(info);
    info->count = parDescr[idx].count;
    info->descr = parDescr[idx].descr;
    info->flags = parDescr[idx].flags;
    info->name = parDescr[idx].name;
    info->type = parDescr[idx].type;
    if (info->data && *dataSize >= (int) cfg->filePathName.size())
        strncpy((char*)info->data, cfg->filePathName.c_str(), *dataSize);
    *dataSize = (int) cfg->filePathName.size() + 1;
    return 0;
}

int RootFilter::filter(FRAMEID frameID, ITEMID instID)
{
    GETINST(instID);
    // get frame type (it can be i16, u32 or double), width and height
    Data_Types type;
    u32 width, height;
    if (mgr->getFrameType(frameID, &type) || mgr->getFrameSize(frameID, &width, &height))
        return -1;

    // direct access to frame buffer (to avoid copy)
    byte *buff = 0;
    u32 size = 0;
    if (mgr->getLockedFrameBuff(frameID, &buff, &size, 100))
        return -1;
    int rc = -1;
    switch (type)
    {
    case TYPE_I16:
        rc = save(frameID, cfg->filePathName.c_str(), (i16 *) buff, width, height);
        break;
    case TYPE_U32:
        rc = save(frameID, cfg->filePathName.c_str(), (u32 *) buff, width, height);
        break;
    case TYPE_DOUBLE:
        rc = save(frameID, cfg->filePathName.c_str(), (double *) buff, width, height);
        break;
    }
    // release frame (it is locked because we have direct access to buffer)
    mgr->unlockFrame(frameID);
    return rc;
}

template<typename T> int RootFilter::save(FRAMEID frameID, const char *fileName, T *buff, u32 width, u32 height)
{
    // here you should save frame data and attributes to your multiframe format
    // all attributes can be obtained like this:
    ExtFrameAttrib attrib;
    ITEMID itemID = 0;
    Data_Types type;
    u32 dataSize = 1024;
    byte *data = new byte[dataSize];
    
    // browse all frame attributes via getFrameFirstAttrib/getFrameNextAttrib
    int rc = mgr->getFrameFirstAttrib(frameID, &itemID, &attrib);
    while (rc == 0)
    {
        if (mgr->getFrameAttrib(frameID, attrib.name, data, &dataSize, &type) == MPXERR_BUFFER_SMALL)
        {
            delete[] data;
            data = new byte[dataSize];
            if (mgr->getFrameAttrib(frameID, attrib.name, data, &dataSize, &type))
                break;
        }
        rc = mgr->getFrameNextAttrib(frameID, &itemID, &attrib);
    }
    delete[] data;
    return 0;
}